module Lazy.Tree exposing
    ( Tree(..), Forest, singleton, build, fromList, fromKeyedList
    , isEmpty, item, children, descendants
    , insert
    , map, map2, filter, filterMap, sort, sortBy, sortWith, stableSortWith, andMap, flatten, andThen, duplicate, extend
    , forestMap, forestMap2
    )

{-| This module implements Rose Tree data structure.

> In computing, a multi-way tree or rose tree is a tree data structure
> with a variable and unbounded number of branches per node.

This particular implementation uses lazy list construction (using `LList` module)
to lazily evaluate levels of Tree.


# Types & Constructor

@docs Tree, Forest, singleton, build, fromList, fromKeyedList


# Query

@docs isEmpty, item, children, descendants


# Modify

@docs insert


# Transforms

@docs map, map2, filter, filterMap, sort, sortBy, sortWith, stableSortWith, andMap, flatten, andThen, duplicate, extend


# Forest

@docs forestMap, forestMap2

-}

import Dict exposing (Dict)
import Lazy
import Lazy.LList as LL exposing (LList)


{-| \*\* Be careful when comparing `Tree`s using `(==)`.\*\*
Due to use of laziness `(==)` isn't reliable for comparing Trees.
-}
type Tree a
    = Tree a (Forest a)


{-| \*\* Be careful when comparing `Forest`s using `(==)`.\*\*
Due to use of laziness `(==)` isn't reliable for comparing Forests.
-}
type alias Forest a =
    LList (Tree a)



-- Tree


{-| Puts value in minimal `Tree` context

    singleton "foo"
        |> item
    --> "foo"

-}
singleton : a -> Tree a
singleton a =
    Tree a LL.empty


{-| Build `Tree` using custom constructor.

This can be for instance used to build `Tree` from other recursive data structure:

    type Item = Item String (List Item)

    getChildren (Item _ children) = children

    Item "foo" [ Item "bar" [], Item "baz" []]
        |> build getChildren
        |> children
    -> [ Item "bar" [], Item "baz" [] ]

Or lookups to some other data structure.

    import Dict exposing (Dict)

    rootItem : String
    rootItem = "foo"

    childrenDict : Dict String (List String)
    childrenDict = Dict.fromList [ ("foo", [ "bar", "baz" ]) ]

    build (\i -> Maybe.withDefault [] <| Dict.get i childrenDict) rootItem
        |> children
    --> [ "bar", "baz" ]

-}
build : (a -> List a) -> a -> Tree a
build getChildren root =
    Tree root <| LL.map (build getChildren) <| LL.llist getChildren root


{-| Check if `Tree` doesn't have any child.

    singleton "foo"
        |> isEmpty
    --> True

    singleton "foo"
        |> insert (singleton "bar")
        |> isEmpty
    --> False

-}
isEmpty : Tree a -> Bool
isEmpty =
    List.isEmpty << children


{-| Obtain item from `Tree`.

    singleton "foo"
        |> item
        |> "foo"

-}
item : Tree a -> a
item (Tree i _) =
    i


{-| Obtain children items of `Tree`.

    singleton "foo"
        |> insert (singleton "bar")
        |> insert (singleton "baz")
        |> children
    --> [ "bar", "baz" ]

-}
children : Tree a -> List a
children =
    List.map item << LL.toList << descendants


{-| Obtain descendants as `Forest` from the `Tree`.

    import Lazy.LList as LL

    singleton "foo"
        |> insert (singleton "bar")
        |> insert (singleton "baz")
        |> descendants
        |> LL.map item
        |> LL.toList
    --> [ "bar", "baz" ]

    singleton "foo"
        |> insert (singleton "bar" |> insert (singleton "baz"))
        |> descendants
        |> LL.map (children)
        |> LL.toList
    --> [ [ "baz" ] ]

-}
descendants : Tree a -> Forest a
descendants (Tree _ d) =
    d


{-| Map function over `Tree`.

    singleton 1
        |> map ((+) 1)
        |> item
    --> 2

    singleton 1
        |> insert (singleton 2)
        |> insert (singleton 3)
        |> map ((*) 2)
        |> children
    --> [ 4, 6 ]

-}
map : (a -> b) -> Tree a -> Tree b
map fc (Tree a forest) =
    Tree (fc a) <| forestMap fc forest


{-| Map function over two `Tree`s

    map2 (+) (singleton 1) (singleton 5)
        |> item
    --> 6

    import Lazy.LList as LL

    Tree 1 (LL.fromList [ singleton 2, singleton 3, singleton 4 ])
        |> map2 (+) (Tree 5 <| LL.fromList [ singleton 6, singleton 7 ])
        |> children
    --> [ 8, 10 ]

-}
map2 : (a -> b -> c) -> Tree a -> Tree b -> Tree c
map2 fc (Tree a1 f1) (Tree a2 f2) =
    Tree (fc a1 a2) <| forestMap2 fc f1 f2


{-| Filter `Tree` children by given function.

This function goes from children of root downwards.
This means that nodes that don't satisfy predicate
are excluded and filter is never performed over their children
even if on those it might pass.

    import Lazy.LList as LL

    Tree 1 (LL.fromList [ singleton 2, singleton 3, singleton 4 ])
        |> filter ((>) 4)
        |> children
    --> [ 2, 3 ]

    Tree 1 (LL.fromList [ insert (singleton 5) <| singleton 2, insert (singleton 6) <| singleton 3, singleton 4 ])
        |> filter ((<) 2)
        |> descendants
        |> LL.map children
        |> LL.toList
    --> [ [ 6 ], [] ]

-}
filter : (a -> Bool) -> Tree a -> Tree a
filter predicate (Tree treeItem c) =
    Tree treeItem <| LL.filterMap (filter_ predicate) c


filter_ : (a -> Bool) -> Tree a -> Maybe (Tree a)
filter_ predicate (Tree treeItem c) =
    if predicate treeItem then
        Just <| Tree treeItem <| LL.filterMap (filter_ predicate) c

    else
        Nothing


{-| FilterMap on `Tree`. Works similarly to `List.filterMap` with the same properties as [filter](#filter).
In case of `filterMap` even root node has to satisfy predicate otherwise
`Nothing` is returned.

    import Lazy.LList as LL

    Tree 1 (LL.fromList [ singleton 2, singleton 3, singleton 4 ])
        |> filterMap (\a -> if a < 4 then Just (a * 2) else Nothing)
        |> Maybe.map children
    --> Just [ 4, 6 ]

    Tree 1 (LL.fromList [ singleton 2, singleton 3, singleton 4 ])
        |> filterMap (\a -> if a > 2 then Just (a * 2) else Nothing)
        |> Maybe.map children
    --> Nothing

-}
filterMap : (a -> Maybe b) -> Tree a -> Maybe (Tree b)
filterMap predicate (Tree treeItem c) =
    predicate treeItem
        |> Maybe.map (\i -> Tree i <| LL.filterMap (filterMap predicate) c)


{-| Sort `Tree`.

    singleton 10
        |> insert (singleton 5)
        |> insert (singleton 2)
        |> sort
        |> children
    --> [ 2, 5 ]

it applies to all levels:

    import Lazy.LList as LL

    singleton 10
        |> insert (Tree 20 <| LL.llist (List.reverse << List.map singleton << List.range 1) 5)
        |> sort
        |> descendants
        |> LL.map children
        |> LL.toList
    --> [ [ 1, 2, 3, 4, 5 ] ]

-}
sort : Tree comparable -> Tree comparable
sort (Tree a f) =
    Tree a <| LL.map sort <| LL.sortBy item f


{-| Sort `Tree` by a function.

    singleton { val = 10 }
       |> insert (singleton { val = 7 })
       |> insert (singleton { val = 3 })
       |> sortBy .val
       |> children
    --> [ { val = 3 }, { val = 7 } ]

it applies to all levels:

    import Lazy.LList as LL

    singleton { a = 10 }
        |> insert (Tree { a = 20 } <| LL.llist (List.reverse << List.map (\v -> singleton { a = v }) << List.range 1) 3)
        |> sortBy .a
        |> descendants
        |> LL.map children
        |> LL.toList
    --> [ [ { a = 1 }, { a = 2 }, { a = 3 } ] ]

-}
sortBy : (a -> comparable) -> Tree a -> Tree a
sortBy predicate (Tree a f) =
    Tree a <|
        LL.map (sortBy predicate) <|
            LL.sortBy (predicate << item) f


{-| Sort `Tree` using custom comparison

    flippedComparison : comparable -> comparable -> Order
    flippedComparison a b =
        case Basics.compare a b of
            LT -> GT
            EQ -> EQ
            GT -> LT

    singleton 10
        |> insert (singleton 2)
        |> insert (singleton 5)
        |> sortWith flippedComparison
        |> children
    --> [ 5, 2 ]

-}
sortWith : (a -> a -> Order) -> Tree a -> Tree a
sortWith compare (Tree a f) =
    Tree a <|
        LL.map (sortWith compare) <|
            LL.sortWith (\fst snd -> compare (item fst) (item snd)) f


{-| Stable sort `Tree` using custom comparison.
The original order is guaranteed to be kept if the comparison returns `EQ`.

    compareAge : { r | age : comparable } -> { r | age : comparable } -> Order
    compareAge a b =
        Basics.compare a.age b.age

    singleton { name = "Eve", age = 55 }
        |> insert (singleton { name = "Joe", age = 25 })
        |> insert (singleton { name = "Sue", age = 25 })
        |> insert (singleton { name = "Johann", age = 23 })
        |> stableSortWith compareAge
        |> children
    --> [ { name = "Johann", age = 23 }, { name = "Joe", age = 25 }, { name = "Sue", age = 25 } ]

-}
stableSortWith : (a -> a -> Order) -> Tree a -> Tree a
stableSortWith compare (Tree a f) =
    Tree a <|
        LL.map (stableSortWith compare) <|
            LL.stableSortWith (\fst snd -> compare (item fst) (item snd)) f


{-| Chain map operations.

    import Lazy.LList as LL
    import Tuple

    Tree Tuple.pair (LL.fromList [ singleton Tuple.pair, singleton Tuple.pair, singleton Tuple.pair ])
        |> andMap (Tree 1 <| LL.fromList [ singleton 2, singleton 3, singleton 4 ])
        |> andMap (Tree 5 <| LL.fromList [ singleton 6, singleton 7 ])
        |> children
    --> [ (2, 6), (3, 7) ]

-}
andMap : Tree a -> Tree (a -> b) -> Tree b
andMap =
    map2 (|>)


{-| Flatten `Tree` of Trees.

    singleton (singleton 1)
        |> flatten
        |> item
    --> 1

    import Lazy.LList as LL

    Tree (Tree "foo" <| LL.fromList [ singleton "bar"]) (LL.fromList [ singleton <| singleton "baz" ])
        |> flatten
        |> children
    --> [ "baz", "bar" ]

-}
flatten : Tree (Tree a) -> Tree a
flatten (Tree (Tree treeItem c) treeChildren) =
    Tree treeItem <| LL.append (LL.map flatten treeChildren) c


{-| Map given function onto a `Tree` and flatten the result.

    import Lazy.LList as LL

    singleton "foo"
        |> insert (singleton "bar")
        |> insert (singleton "baz")
        |> andThen (\a -> Tree a <| LL.fromList [ singleton <| a ++ " fighter" ])
        |> children
    --> [ "bar", "baz", "foo fighter" ]

-}
andThen : (a -> Tree b) -> Tree a -> Tree b
andThen fc =
    flatten << map fc


{-| Insert one `Tree` as children another.

    singleton 1
        |> insert (singleton 2)
        |> insert (singleton 3)
        |> children
    --> [ 2, 3 ]

    singleton 1
        |> insert (singleton 2)
        |> item
    --> 1

-}
insert : Tree a -> Tree a -> Tree a
insert t (Tree treeItem c) =
    Tree treeItem <| LL.append c <| LL.fromList [ t ]



-- Forest


{-| Construct `Forest` from a list.

    import Lazy.LList as LL

    [ { id = 1, parent = Nothing }
    , { id = 2, parent = Nothing }
    , { id = 3, parent = Just 1 }
    ]
        |> fromList (\p i -> Maybe.map .id p == i.parent)
        |> LL.map (.id << item)
        |> LL.toList
    --> [ 1, 2 ]

    [ { id = 1, parent = Nothing }
    , { id = 2, parent = Nothing }
    , { id = 3, parent = Just 1 }
    , { id = 4, parent = Just 1 }
    , { id = 5, parent = Just 2 }
    ]
        |> fromList (\p i -> Maybe.map .id p == i.parent)
        |> LL.andThen descendants
        |> LL.map (.id << item)
        |> LL.toList
    --> [ 3, 4, 5 ]

-}
fromList : (Maybe a -> a -> Bool) -> List a -> Forest a
fromList isParent =
    fromList_ Nothing isParent


fromList_ : Maybe a -> (Maybe a -> a -> Bool) -> List a -> Forest a
fromList_ parent isParent list =
    LL.llist (List.map (constructTree isParent list) << List.filter (isParent parent)) list


{-| Construct `Forest` from a list of items that can be **uniquely identified** by comparable value.

This function can yield much better performance than more general [`fromList`](#fromList) alternative.
Be aware that this function pushes **more work on initial construction of the Tree** at the benefit of
doing more optimal operations during evaluation of the Tree at later time.

    import Lazy.LList as LL

    [ { id = 1, parent = Nothing }
    , { id = 2, parent = Nothing }
    , { id = 3, parent = Just 1 }
    ]
        |> fromKeyedList .id (Maybe.withDefault [] << Maybe.map List.singleton << .parent)
        |> LL.map (.id << item)
        |> LL.toList
    --> [ 1, 2 ]

    [ { id = 1, parent = Nothing }
    , { id = 2, parent = Nothing }
    , { id = 3, parent = Just 1 }
    , { id = 4, parent = Just 1 }
    , { id = 5, parent = Just 2 }
    ]
        |> fromList (\p i -> Maybe.map .id p == i.parent)
        |> LL.andThen descendants
        |> LL.map (.id << item)
        |> LL.toList
    --> [ 3, 4, 5 ]

-}
fromKeyedList : (a -> comparable) -> (a -> List comparable) -> List a -> Forest a
fromKeyedList getNodeId getParentIds list =
    let
        -- Split out list of roots (nodes with parent Nothing)
        -- and Dict which maps parentIds to list of children
        -- This means you have to pay `O(n*log(n))` once, when you call this function
        -- but then each subsequent level expansions will be cheaper (`O(log(n))` lookup in Dict)
        ( roots, parentIdToChildren ) =
            List.foldr
                (\a ( rts, ptch ) ->
                    case getParentIds a of
                        [] ->
                            -- a is root, add it to list of roots
                            ( a :: rts
                            , ptch
                            )

                        parentIds ->
                            -- a is non-root - prepend it to the list of children under the id of its parent
                            ( rts
                            , List.foldr
                                (\parentId acc ->
                                    Dict.update parentId
                                        (\mval ->
                                            case mval of
                                                Nothing ->
                                                    Just [ a ]

                                                Just kids ->
                                                    Just (a :: kids)
                                        )
                                        acc
                                )
                                ptch
                                parentIds
                            )
                )
                ( [], Dict.empty )
                list

        buildTree : a -> Tree a
        buildTree =
            build
                (\a ->
                    -- This is the `O(log(n))` work done when expanding level
                    Dict.get (getNodeId a) parentIdToChildren |> Maybe.withDefault []
                )
    in
    LL.map buildTree <| LL.fromList roots


{-| Map function over `Forest`.

    import Lazy.LList as LL

    [ 1, 2, 3 ]
        |> fromList (\m _ -> m == Nothing)
        |> forestMap ((+) 1)
        |> LL.map item
        |> LL.toList
    --> [ 2, 3, 4 ]

-}
forestMap : (a -> b) -> Forest a -> Forest b
forestMap fc =
    LL.map (map fc)


{-| Map function over two `Forest`s.

    import Lazy.LList as LL

    [ 1, 2, 3 ]
        |> fromList (\m _ -> m == Nothing)
        |> forestMap2 (+) (fromList (\m _ -> m == Nothing) [1, 2])
        |> LL.map item
        |> LL.toList
    --> [ 2, 4 ]

-}
forestMap2 : (a -> b -> c) -> Forest a -> Forest b -> Forest c
forestMap2 fc =
    LL.map2 (map2 fc)


{-| Duplicates Tree (Comonad)
-}
duplicate : Tree a -> Tree (Tree a)
duplicate ((Tree _ xs) as t) =
    Tree t <| LL.map duplicate xs


{-| Extend tree (Comonad)
-}
extend : (Tree a -> b) -> Tree a -> Tree b
extend f =
    map f << duplicate



-- Private


constructTree : (Maybe a -> a -> Bool) -> List a -> a -> Tree a
constructTree isParent list treeItem =
    Tree treeItem <| fromList_ (Just treeItem) isParent list
